//
// Created by kirdyk on 13.09.16.
//

#ifndef WOAP2_ICHECK_H
#define WOAP2_ICHECK_H

#include <string>
#include <vector>
#include <sstream>
#include <map>

#include "../logger.h"
#include "../constants.h"

namespace checks
{
    /**
     * Parameters map type.
     */
    typedef std::map<std::string, std::string> ParamsMap;
    
    /**
     * Check parameter description.
     */
    struct CheckParameter
    {
        /**
         * Parameter name.
         */
        std::string name;
        /**
         * Parameter type.
         */
        std::string type;
        /**
         * Parameter length.
         */
        int length;
    
        /**
         * Default value
         */
        std::string default_value;
    };
    
    /**
     * Check result structure.
     */
    struct CheckResult
    {
        /**
         * Success flag.
         */
        bool success;
        /**
         * Limit exceede flag.
         */
        bool limit_exceeded;
        /**
         * Result string
         */
        std::string result;
        /**
         * Error message
         */
        std::string error_message;
    };
    
    /**
     * Check base class.
     */
    class CheckBase
    {
    protected:
        /**
         * Check parameters list.
         */
        std::vector<CheckParameter> check_params;
    
        /***
         * Combo parameters map type
         */
        typedef std::map<std::string, std::vector<std::string> > ComboMap;
    
        /**
         * combo values
         */
        ComboMap combo_values_map;
    public:
    
        /***
         * Constructor
         */
        CheckBase()
        {
            WOAP2_LOG_TRACE(" > CheckBase::CheckBase")
            WOAP2_LOG_TRACE(" < CheckBase::CheckBase")
        }
    
        /***
         * Destructor
         */
        virtual ~CheckBase()
        {
            WOAP2_LOG_TRACE(" > CheckBase::~CheckBase")
            WOAP2_LOG_TRACE(" < CheckBase::~CheckBase")
        }
        
        /**
         * Get check description.
         * @return check description.
         */
        virtual std::string get_description() = 0;
    
        /**
         * Execute check.
         * @param params check parameters.
         * @return check result.
         */
        virtual CheckResult execute(ParamsMap const &params) = 0;
    
        /**
         * Get check parameters definition.
         * @return parameters definition string.
         */
        std::string get_params_definition()
        {
            WOAP2_LOG_TRACE(" > CheckBase::get_params_definition")
   
            std::ostringstream out;
            
            out << "#";
            
            int param_num = 1;
            
            for(std::vector<CheckParameter>::const_iterator it = check_params.begin();
                it != check_params.end(); ++it, ++param_num)
            {
                out << "paramname" << param_num << "=" << it->name << "#"
                    << "paramtype" << param_num << "=" << it->type << "#";
                if(it->length > 0)
                {
                    out << "maxlength" << param_num << "=" << it->length << "#";
                }
    
                if(!it->default_value.empty())
                {
                    out << "default" << param_num << "=" << it->default_value << "#";
                }
                
                ComboMap::const_iterator combo_values = combo_values_map.find(it->name);
    
                if(combo_values != combo_values_map.end() && combo_values->second.size() > 0)
                {
                    out << "valuecount" << "=" << combo_values->second.size() << "#";
    
                    int combo_param_num = 1;
        
                    for(std::vector<std::string>::const_iterator it = combo_values->second.begin();
                        it != combo_values->second.end(); ++it, ++combo_param_num)
                    {
                        out << "value" << combo_param_num << "=" << (*it) << "#";
                    }
        
                    out << "atbutton_paramcount=1#";
                    out << "atbutton_param1=" << it->name << "#";
                }
    
            }
    
            WOAP2_LOG_TRACE(" < CheckBase::get_params_definition")
            return out.str();
        }
    };
    
    /***
     * Params accessor empty template
     */
    template<typename T>
    struct params_accessor
    {
    };
    
    /***
     * String parameters accessor
     */
    template<>
    struct params_accessor<std::string>
    {
        /***
         * Get optional parameter from the map.
         * @param params parameters map
         * @param param_name parameter's name
         * @param default_value defaukt value
         * @return value of the parameter
         */
        std::string get_optional_param(ParamsMap const &params, const std::string &param_name,
                                       const std::string &default_value = "")
        {
            ParamsMap::const_iterator param_it = params.find(param_name);
            if(param_it == params.end())
            {
                return default_value;
            }
            
            return param_it->second;
        }
    
        /***
         * Get required parameter from the map or throw an exception otherwise
         * @param params parameters map
         * @param param_name parameter's name
         * @param error_header exception message header
         * @return value of the parameter
         */
        std::string
        get_required_param(ParamsMap const &params, const std::string &param_name, const std::string &error_header)
        {
            ParamsMap::const_iterator param_it = params.find(param_name);
            if(param_it == params.end())
            {
                WOAP2_LOG_ERROR(error_header + param_name + constants::param_missing_string);
                
                throw std::runtime_error(constants::no_param_error_string);
            }
            
            return param_it->second;
        }
    };
    
    /***
     * Integer parameters accessor
     */
    template<>
    struct params_accessor<int>
    {
        /***
         * Get optional parameter from the map.
         * @param params parameters map
         * @param param_name parameter's name
         * @param default_value defaukt value
         * @return value of the parameter
         */
        int get_optional_param(ParamsMap const &params, const std::string &param_name, int default_value = 0)
        {
            ParamsMap::const_iterator param_it = params.find(param_name);
            if(param_it == params.end())
            {
                return default_value;
            }
            
            return std::stoi(param_it->second);
        }
    
        /***
         * Get required parameter from the map or throw an exception otherwise
         * @param params parameters map
         * @param param_name parameter's name
         * @param error_header exception message header
         * @return value of the parameter
         */
        int get_required_param(ParamsMap const &params, const std::string &param_name, const std::string &error_header)
        {
            ParamsMap::const_iterator param_it = params.find(param_name);
            if(param_it == params.end())
            {
                WOAP2_LOG_ERROR(error_header + param_name + constants::param_missing_string);
                
                throw std::runtime_error(constants::no_param_error_string);
            }
            
            return std::stoi(param_it->second);
        }
    };
    
    /***
     * Boolean parameters accessor
     */
    template<>
    struct params_accessor<bool>
    {
        /***
         * Get optional parameter from the map.
         * @param params parameters map
         * @param param_name parameter's name
         * @param default_value defaukt value
         * @return value of the parameter
         */
        bool get_optional_param(ParamsMap const &params, const std::string &param_name, bool default_value = false)
        {
            ParamsMap::const_iterator param_it = params.find(param_name);
            if(param_it == params.end())
            {
                return default_value;
            }
            
            return param_it->second == "true";
        }
    
        /***
         * Get required parameter from the map or throw an exception otherwise
         * @param params parameters map
         * @param param_name parameter's name
         * @param error_header exception message header
         * @return value of the parameter
         */
        bool get_required_param(ParamsMap const &params, const std::string &param_name, const std::string &error_header)
        {
            ParamsMap::const_iterator param_it = params.find(param_name);
            if(param_it == params.end())
            {
                WOAP2_LOG_ERROR(error_header + param_name + constants::param_missing_string);
                
                throw std::runtime_error(constants::no_param_error_string);
            }
            
            return param_it->second == "true";
        }
    };
    
}


#endif //WOAP2_ICHECK_H
