//
// Created by bezgin on 11.09.16.
//

#ifndef WOAP2_CHECKS_H
#define WOAP2_CHECKS_H

#include <string>
#include <map>
#include <memory>
#include "../utils.h"
#include "../programoption.h"

#include "check.h"

#include "cpucheck.h"
#include "diskspacecheck.h"
#include "memorycheck.h"
#include "proccheck.h"
#include "pingcheck.h"
#include "certcheck.h"
#include "urlcheck.h"
#include "snmpcheck.h"
#include "../logger.h"

namespace checks
{
    /**
     * Unknow check class.
     */
    class UnknownCheck : public CheckBase
    {
    public:
        /**
         * Constructor. Just empty.
         */
        UnknownCheck()
        {
            WOAP2_LOG_TRACE(" > UnknownCheck::UnknownCheck")
            WOAP2_LOG_DEBUG("UnknownCheck")
            WOAP2_LOG_TRACE(" < UnknownCheck::UnknownCheck")
        }
    
        /***
         * Destructor
         */
        virtual ~UnknownCheck()
        {
            WOAP2_LOG_TRACE(" > UnknownCheck::~UnknownCheck")
            WOAP2_LOG_TRACE(" < UnknownCheck::~UnknownCheck")
        }
        
        /**
         * Execute check.
         * @param params check parameters.
         * @return "Unknown check" error result.
         */
        virtual CheckResult execute(ParamsMap const &params) override
        {
            WOAP2_LOG_TRACE(" > UnknownCheck::execute")
            WOAP2_LOG_DEBUG("Executing UnknownCheck.")
            WOAP2_LOG_TRACE(" < UnknownCheck::execute")
            
            return CheckResult{false, false, "", "Unknown check"};
        }
    
        /**
         * Get check name.
         * @return "Unknown" as check name.
         */
        static std::string get_name()
        {
            return "Unknown";
        }
    
        /**
         * Get check description
         * @return "Unknown check" is the description.
         */
        virtual std::string get_description() override
        {
            return "Unknown check";
        }
    };
    
    /**
     * Check factory class.
     */
    class CheckFactory
    {
        /***
         * Check creation function template
         * @param for_params Just fill parameters description
         * @return check object
         */
        template<class T>
        static std::unique_ptr<CheckBase> creator_func(bool for_params)
        {
            return std::unique_ptr<CheckBase>(new T(for_params));
        }
    
        /**
         * Check creator function type.
         */
        typedef std::unique_ptr<CheckBase>(*CheckCreatorFunc)(bool);
    
        /**
         * Check creators map type.
         */
        typedef std::map<std::string, CheckCreatorFunc> ChecksMap;
    
        /**
         * Check parameter type.
         */
        typedef std::pair<std::string, std::string> ParamsListEntry;
        /**
         * Check parameters list type.
         */
        typedef std::vector<ParamsListEntry> ParamsList;
    
        /**
         * Check creators map.
         */
        ChecksMap checks_map;
    
        template<class T>
        void add_check(std::shared_ptr<Settings::allowed_checks_type> allowed_checks)
        {
            if(allowed_checks->find(T::get_name()) != allowed_checks->end())
            {
                checks_map[T::get_name()] = creator_func<T>;
            }
        }
        
    public:
    
        CheckFactory()
        {
            WOAP2_LOG_TRACE(" > CheckFactory::CheckFactory/0")
            WOAP2_LOG_TRACE(" < CheckFactory::CheckFactory/0")
        }
        
        /**
         * Constructor. Fills check creators map.
         */
        CheckFactory(Settings &settings)
        {
            WOAP2_LOG_TRACE(" > CheckFactory::CheckFactory")
    
            std::shared_ptr<Settings::allowed_checks_type> allowed_checks = settings.get_checks_set();
    
            add_check<CpuCheck>(allowed_checks);
            add_check<DiskSpaceCheck>(allowed_checks);
            add_check<RamCheck>(allowed_checks);
            add_check<SwapCheck>(allowed_checks);
            add_check<ProcCheck>(allowed_checks);
            add_check<PingCheck>(allowed_checks);
            add_check<CertificateCheck>(allowed_checks);
            add_check<UrlCheck>(allowed_checks);
            add_check<SnmpCheck>(allowed_checks);
    
            WOAP2_LOG_TRACE(" < CheckFactory::CheckFactory")
        }
    
        /**
         * Get check object for the specified check name.
         * @param check_name the name of the check.
         * @return check object.
         */
        std::unique_ptr<CheckBase> get_check(std::string const &check_name, bool for_params) const
        {
            WOAP2_LOG_TRACE(" > CheckFactory::get_check")
            WOAP2_LOG_DEBUG(std::string("Getting check  ") + check_name)
            
            ChecksMap::const_iterator check_it = checks_map.find(check_name);
    
            // If there is no such check, return unknown check object.
            if(check_it == checks_map.end())
            {
                WOAP2_LOG_ERROR(std::string("Unknown check ") + check_name)
                WOAP2_LOG_TRACE(" < CheckFactory::get_check")
                
                return std::unique_ptr<CheckBase>(new UnknownCheck());
            }
    
            WOAP2_LOG_TRACE(" < CheckFactory::get_check")
            return check_it->second(for_params);
        }
    
        /**
         * List checks available
         * @return the list of the available checks and their count.
         */
        std::pair<int, std::string> list_checks() const
        {
            WOAP2_LOG_TRACE(" > CheckFactory::list_checks")
            
            int checks_count = checks_map.size();
            int check_num = 1;
            
            ParamsList params_list(checks_count * 3);
            
            for(ChecksMap::const_iterator check_it = checks_map.begin();
                check_it != checks_map.end(); check_it++, check_num++)
            {
                std::unique_ptr<CheckBase> check = check_it->second(false);
                const std::string &check_num_str = std::to_string(check_num);
                
                params_list[(check_num - 1) * 3] = ParamsListEntry("checkid" + check_num_str, check_it->first);
                params_list[(check_num - 1) * 3 + 1] = ParamsListEntry("checkdescription" + check_num_str,
                                                                       check->get_description());
                params_list[(check_num - 1) * 3 + 2] = ParamsListEntry("available" + check_num_str, "true");
            }
    
            WOAP2_LOG_TRACE(" < CheckFactory::list_checks")
            return std::pair<int, std::string>(checks_count, utils::params_list_to_string(params_list));
        }
    };
}

#endif //WOAP2_CHECKS_H
