//
// Created by kirdyk on 14.10.16.
//

#include "pingcheck.h"
#include <memory>
#include <cstdio>
#include "../utils.h"

namespace checks
{
    static const std::string pingcheck_error_string = "Ping check Error: ";
    
    double PingCheck::perform_ping(std::string &host, bool ipv6) const
    {
        WOAP2_LOG_TRACE(" > PingCheck::perform_ping");
        
        std::string cmd = (ipv6 ? "ping6 -c 1 \"" : "ping -c 1 \"") + host + "\" 2>&1";
        WOAP2_LOG_TRACE(std::string("Executing ") + cmd)
        
        // Create a process
        auto proc = std::shared_ptr<FILE>(popen(cmd.c_str(), "r"), pclose);
        
        char buf[256];
        boost::regex whitespace("\\s+");
        
        std::string first_line = "Error executing ping.";
        
        // read first line
        if(fgets(buf, 255, proc.get()))
        {
            first_line = utils::trim(std::string(buf));
        }
        
        bool stats_found = false;
        // find the statictics line in the output
        while(!stats_found && fgets(buf, 255, proc.get()))
        {
            std::string data_string(buf);
            
            if(data_string.find("ping statistics") != std::string::npos)
            {
                stats_found = true;
            }
        }
    
        // if there is no stats -- ping failed and reason is in the first line of output
        if(!stats_found)
        {
            WOAP2_LOG_WARN("ping failed: " + first_line);
            WOAP2_LOG_TRACE(" < PingCheck::perform_ping");
            return -1;
        }
    
        // statistics line parsing RE
        boost::regex stats_re("((?:[\\w-]+\\/)+(?:[\\w-]+\\s))[^\\d\\.]+((?:[\\d\\.]+\\/)+(?:[\\d\\.]+\\s))");
        boost::regex slash_re("\\/");
        // packets loss parsing RE
        boost::regex packet_loss_re("(\\d{1,3}(?:.\\d+)?)%\\s+packet loss");
        
        stats_found = false;
        double avg_rtt = -1;
        double loss_percent = 0;
    
        // search the output for rtt and/or loss
        while(!stats_found && fgets(buf, 255, proc.get()))
        {
            std::string data_string(buf);
            
            boost::match_results<std::string::const_iterator> re_match;
    
            // check if output line contains the loss info
            if(boost::regex_search(data_string.cbegin(), data_string.cend(), re_match, packet_loss_re))
            {
                std::string packet_loss_str(re_match[1].first, re_match[1].second);
                
                loss_percent = std::stod(packet_loss_str);
                
                continue;
            }
    
            // check if output line contains the timing info
            if(boost::regex_search(data_string.cbegin(), data_string.cend(), re_match, stats_re))
            {
                if(re_match.size() > 2)
                {
                    stats_found = true;
                    
                    std::string stats_names_str(re_match[1].first, re_match[1].second);
                    std::string stats_vals_str(re_match[2].first, re_match[2].second);
                    
                    std::vector<std::string> stats_names = utils::split(stats_names_str, slash_re);
                    std::vector<std::string> stats_vals = utils::split(stats_vals_str, slash_re);
                    
                    int i = 0;
                    
                    for(std::vector<std::string>::const_iterator it = stats_names.cbegin();
                        it != stats_names.cend(); ++it, ++i)
                    {
                        if(*it == "avg")
                        {
                            break;
                        }
                    }
                    
                    std::string avg_str = stats_vals[i];
                    avg_rtt = std::stod(avg_str);
                }
            }
        }
    
        WOAP2_LOG_TRACE("avg_rtt=" + utils::to_string(avg_rtt) + "ms; loss=" + utils::to_string(loss_percent) + "%");
    
        WOAP2_LOG_TRACE(" < PingCheck::perform_ping");
        return (loss_percent != 0) ? -1 : avg_rtt;
    }
    
    CheckResult PingCheck::execute(ParamsMap const &params)
    {
        WOAP2_LOG_TRACE(" > PingCheck::execute")
        try
        {
            // Get parameters
            auto stringParamsAccessor = params_accessor<std::string>();
            std::string host = stringParamsAccessor.get_required_param(params, host_param_name, pingcheck_error_string);
    
            auto boolParamsAccessor = params_accessor<bool>();
            bool useIPv6 = boolParamsAccessor.get_optional_param(params, ipv6_param_name, false);
    
            auto intParamsAccessor = params_accessor<int>();
            int packets_to_send = intParamsAccessor.get_required_param(params, pkts_num_param_name,
                                                                       pingcheck_error_string);
    
            // We have total timeout and number of packets. Calculate avg rtt limit for the individual packet.
            int total_timeout = intParamsAccessor.get_required_param(params, timeout_param_name,
                                                                     pingcheck_error_string);
            double rtt_limit = total_timeout * 1000.0 / packets_to_send;
            int limit_rate = intParamsAccessor.get_required_param(params, limit_param_name, pingcheck_error_string);
    
            WOAP2_LOG_DEBUG(std::string("Pinging ") + host + " " +
                            std::to_string(packets_to_send) + " times for " +
                            std::to_string(total_timeout) + " s. (" + utils::to_string(rtt_limit) + " ms per try)" +
                            " with a limit " + std::to_string(limit_rate) + "%");
    
            // number of packets with rtt <= limit
            int under_limit_count = 0;
            // number of success pings
            int success_count = 0;
            double rtt_acc = 0;
    
            // Do some pinging
            for(int i = 0; i < packets_to_send; ++i)
            {
                double rtt = perform_ping(host, useIPv6);
    
                if(rtt > 0)
                {
                    success_count++;
                    rtt_acc += rtt;
                    if(rtt <= rtt_limit)
                    {
                        under_limit_count++;
                    }
                }
            }
    
            // calculate the avg rtt
            double avg_rtt = rtt_acc / ((double) success_count);
    
            WOAP2_LOG_TRACE(std::to_string(success_count) + " pings succedded, " +
                            std::to_string(under_limit_count) + "under limit");
    
            // calculate the rate of the packets which fall withing the limit
            int under_limit_rate = (under_limit_count * 100) / packets_to_send;
            bool under_limit = (under_limit_rate >= limit_rate);
    
            std::string result = utils::to_string(under_limit_rate) + "% " +
                                 (under_limit ? "" : " out of Limit " + std::to_string(limit_rate) + "%") +
                                 " pings fall into rtt limit. Avg. rtt is " + utils::to_string(avg_rtt);
    
            WOAP2_LOG_TRACE(" < PingCheck::execute")
            return CheckResult{true, under_limit, result, ""};
        }
        catch(std::runtime_error &exception)
        {
            std::string err_string = exception.what();
            WOAP2_LOG_TRACE(" < PingCheck::execute")
            return CheckResult{false, false, "", err_string};;
        }
        catch(...)
        {
            WOAP2_LOG_ERROR("Unexpected exception");
            WOAP2_LOG_TRACE(" < PingCheck::execute")
            return CheckResult{false, false, "", "Unexpected exception"};
        }
    }
}