//
// Created by bezgin on 30.11.16.
//

#include "urlcheck.h"
#include <boost/asio.hpp>
#include "../httpclient.h"
#include "../utils.h"

namespace checks
{
    /***
     * URL check error header string
     */
    static const std::string urlcheck_error_string = "URL check Error: ";
    
    CheckResult UrlCheck::execute(ParamsMap const &params)
    {
        WOAP2_LOG_TRACE(" > UrlCheck::execute")
        try
        {
            // Check if required parameters are present.
            auto stringParamAccessor = params_accessor<std::string>();
            std::string url = stringParamAccessor.get_required_param(params, url_param_name, urlcheck_error_string);
    
            auto intParamsAccessor = params_accessor<int>();
            unsigned int status = intParamsAccessor.get_optional_param(params, status_param_name, 0);
            std::string search_text = stringParamAccessor.get_optional_param(params, search_text_param_name, "");
    
            // defaults
            bool use_ssl = false;
            std::string host = url, port = "80", path = "/";
            std::string url_original = url;
    
            // find protocol part of the URL
            // If it is https, use ssl and default to 443 port
            size_t proto_pos = url.find("://");
            if(proto_pos != std::string::npos)
            {
                std::string proto = url.substr(0, proto_pos);
                std::transform(proto.begin(), proto.end(), proto.begin(), ::tolower);
        
                if(proto == "https")
                {
                    use_ssl = true;
                    port = "443";
                }
        
                url = url.substr(proto_pos + 3);
            }
    
            // find path part of the URL
            size_t path_pos = url.find("/");
    
            if(path_pos != std::string::npos)
            {
                path = url.substr(path_pos);
                host = url.substr(0, path_pos);
            }
            else
            {
                host = url;
            }
    
            // find port in the host part
            if(host.find(':') != std::string::npos)
            {
                boost::regex colon_re(":");
                std::vector<std::string> host_parts = utils::split(host, colon_re);
    
                if(host_parts.size() > 1)
                {
                    host = host_parts[0];
                    port = host_parts[1];
                }
            }
    
            WOAP2_LOG_DEBUG(std::string("Checking ") + (use_ssl ? "HTTPS " : "HTTP ") + std::string("URL ") +
                            url_original + std::string(" for ") +
                            (status == 0 ? "non-fail" : std::to_string(status)) + std::string(" status"))
    
            // Do the actual test
            boost::asio::io_service io_service;
            std::unique_ptr<http::HttpClientBase> http_client;
    
            // Choose the correct client implementation
            if(!use_ssl)
            {
                http_client = std::unique_ptr<http::HttpClientBase>(
                        new http::HttpClient(io_service, host, port, path, search_text));
            }
            else
            {
                http_client = std::unique_ptr<http::HttpClientBase>(
                        new http::HttpsClient(io_service, host, port, path, search_text));
            }
            // Run boost::asio event loop
            io_service.run();
    
            // URL status code, status message and failed flag
            unsigned int url_status = http_client->get_status();
            std::string status_msg = http_client->get_status_message();
            bool is_failed = http_client->is_failed();
    
            WOAP2_LOG_TRACE(std::string("Check finished. ") +
                            (!is_failed
                             ? "Code is " +
                               std::to_string(url_status) +
                               std::string(" (") + status_msg + std::string(")")
                             : "FAILED: " + http_client->get_message()))
    
            // Check if URL status is OK for us
            bool status_ok = (status == 0) ? url_status == 200 : url_status == status;
            bool need_text_search = http_client->need_to_search();
            bool text_found = http_client->is_text_found();
    
            WOAP2_LOG_TRACE(" < UrlCheck::execute")
    
            return CheckResult{
                    !is_failed,
                    status_ok && (!need_text_search || text_found),
                    !is_failed
                    ? std::to_string(url_status) + " " + status_msg +
                      ((!text_found && need_text_search) ? " Text not found " : "")
                    : "",
                    is_failed ? http_client->get_message() : ""};
        }
        catch(boost::system::system_error error)
        {
            WOAP2_LOG_ERROR(error.code().message());
            WOAP2_LOG_TRACE(" < UrlCheck::execute");
            return CheckResult{false, false, "", error.code().message()};
        }
        catch(std::runtime_error error)
        {
            WOAP2_LOG_ERROR(error.what());
            WOAP2_LOG_TRACE(" < UrlCheck::execute");
            return CheckResult{false, false, "", error.what()};
        }
        catch(...)
        {
            WOAP2_LOG_ERROR("Unexpected exception");
            WOAP2_LOG_TRACE(" < UrlCheck::execute");
            return CheckResult{false, false, "", "Unexpected exception"};
        }
    }
}