//
// Created by bezgin on 11.09.16.
//

#ifndef WOAP2_COMMANDS_H
#define WOAP2_COMMANDS_H

#include <memory>
#include <string>
#include <map>

#include "utils.h"
#include "checks/checks.h"
#include "programoption.h"
#include "logger.h"

namespace commands
{
    /**
     * Parameters map type.
     */
    typedef std::map<std::string, std::string> ParamsMap;
    
    /**
     * Command base class
     */
    struct CommandBase
    {
        /**
         * Constructor. Creates new checks factory.
         */
        CommandBase()
                : checks_factory(checks::CheckFactory())
        {
            WOAP2_LOG_TRACE(" > CommandBase::CommandBase/0")
            WOAP2_LOG_TRACE(" < CommandBase::CommandBase/0")
        }
    
        /**
         * Constructor. Uses specified checks factory
         * @param _checks_factory existing checks factory.
         */
        CommandBase(checks::CheckFactory const &_checks_factory)
                : checks_factory(_checks_factory)
        {
            WOAP2_LOG_TRACE(" > CommandBase::CommandBase/1")
            WOAP2_LOG_TRACE(" < CommandBase::CommandBase/1")
        }
    
        virtual ~CommandBase()
        {
            WOAP2_LOG_TRACE(" > CommandBase::~CommandBase")
            WOAP2_LOG_TRACE(" < CommandBase::~CommandBase")
        }
        
        /**
         * Execute the command.
         * @return result of the execution.
         */
        virtual std::string execute() = 0;

    protected:
        /**
         * Checks factory.
         */
        checks::CheckFactory const &checks_factory;
    
    };
    
    /**
     * Error command.
     */
    class ErrorCommand : public CommandBase
    {
    private:
        /**
         * Error message.
         */
        std::string message;
    
    public:
    
        /**
         * Constructor.
         * @param message error message.
         */
        ErrorCommand(std::string message)
        {
            WOAP2_LOG_TRACE(" > ErrorCommand::ErrorCommand/1")
            WOAP2_LOG_DEBUG("ErrorCommand")
            this->message = message;
            WOAP2_LOG_TRACE(" < ErrorCommand::ErrorCommand/1")
        }
    
        virtual ~ErrorCommand()
        {
            WOAP2_LOG_TRACE(" > ErrorCommand::~ErrorCommand")
            WOAP2_LOG_TRACE(" < ErrorCommand::~ErrorCommand")
        }
        
        /**
         * Execute the command.
         * @return error message.
         */
        virtual std::string execute()
        {
            WOAP2_LOG_TRACE(" > ErrorCommand::execute")
            WOAP2_LOG_DEBUG("Executing ErrorCommand.")
            WOAP2_LOG_TRACE(" < ErrorCommand::execute")
            return message;
        }
    };
    
    /**
     * Unknow command error command.
     */
    class UnknownCommand : public CommandBase
    {
        /**
         * Command string.
         */
        std::string cmd;
    
    public:
        /**
         * Constructor
         * @param command the name of unknown command.
         */
        UnknownCommand(std::string command)
        {
            WOAP2_LOG_TRACE(" > UnknownCommand::UnknownCommand/1")
            WOAP2_LOG_DEBUG("UnknownCommand")
            
            this->cmd = command;
    
            WOAP2_LOG_TRACE(" < UnknownCommand::UnknownCommand/1")
        }
    
        virtual ~UnknownCommand()
        {
            WOAP2_LOG_TRACE(" > UnknownCommand::~UnknownCommand")
            WOAP2_LOG_TRACE(" < UnknownCommand::~UnknownCommand")
        }
        
        /**
         * Execute the command.
         * @return error message.
         */
        virtual std::string execute()
        {
            WOAP2_LOG_TRACE(" > UnknownCommand::execute")
            WOAP2_LOG_DEBUG(std::string("Executing UnknownCommand. cmd=") + cmd)
            WOAP2_LOG_TRACE(" < UnknownCommand::execute")
            
            return "Command " + cmd + " not implemented\n";
        }
    };
    
    /**
     * Run check command.
     */
    class RunCheckCommand : public CommandBase
    {
        /**
         * ID of the check.
         */
        std::string check_id;
        /**
         * Check parameters.
         */
        std::string check_params;

    public:
        /**
         * Constructor. Extracts check name and parameters for futher usage.
         * @param params command parameters map.
         * @param _checks_factory existing checks factory.
         */
        RunCheckCommand(const ParamsMap &params, checks::CheckFactory const &_checks_factory)
                : CommandBase(_checks_factory)
        {
            WOAP2_LOG_TRACE(" > RunCheckCommand::RunCheckCommand/2")
            
            ParamsMap::const_iterator check_it = params.find("checkid");
    
            if(check_it != params.end())
            {
                check_id = check_it->second;
            }
    
            ParamsMap::const_iterator para_it = params.find("para");
    
            if(para_it != params.end())
            {
                check_params = para_it->second.substr(1, para_it->second.length() - 2);
            }
    
            WOAP2_LOG_DEBUG(std::string("RunCheckCommand with checkid=") + check_id + " and para=" + check_params);
    
            WOAP2_LOG_TRACE(" < RunCheckCommand::RunCheckCommand/2")
        }
    
        virtual ~RunCheckCommand()
        {
            WOAP2_LOG_TRACE(" > RunCheckCommand::~RunCheckCommand")
            WOAP2_LOG_TRACE(" < RunCheckCommand::~RunCheckCommand")
        }
        
        /**
         * Execute run check.
         * @return check result string
         */
        virtual std::string execute()
        {
            WOAP2_LOG_TRACE(" > RunCheckCommand::execute")
            WOAP2_LOG_DEBUG(std::string("Executing RunCheckCommand. checkid=") + check_id)
            
            if(check_id.length() == 0)
            {
                WOAP2_LOG_WARN("RunCheckCommand::execute: empty check_id")
                WOAP2_LOG_TRACE(" < RunCheckCommand::execute")
                
                return "#keycount=4#responseaction=runcheck#sessionid=1#checkid=" +
                       check_id + "#errormessage=Check not found#";
            }
    
            std::unique_ptr<checks::CheckBase> check = checks_factory.get_check(check_id, false);
    
            checks::CheckResult result = check->execute(utils::get_params_map(check_params));
    
            if(!result.success)
            {
                WOAP2_LOG_TRACE(" < RunCheckCommand::execute")
                return "#keycount=4#responseaction=runcheck#sessionid=1#checkid=" +
                       check_id + "#errormessage=" + result.error_message + "#";
            }
            else
            {
                WOAP2_LOG_TRACE(" < RunCheckCommand::execute")
                return "#keycount=5#responseaction=runcheck#sessionid=1#checkid=" +
                       check_id +
                       "#result=" + (result.limit_exceeded ? "up" : "down") +
                       "#resultstring=" + result.result + "#";
            }
        }
    };
    
    /**
     * List available checks command
     */
    class ListChecksCommand : public CommandBase
    {

    public:
        
        /**
         * Constructor
         * @param _checks_factory existing checks factory.
         */
        ListChecksCommand(const ParamsMap &params, checks::CheckFactory const &_checks_factory)
                : CommandBase(_checks_factory)
        {
            WOAP2_LOG_TRACE(" > ListChecksCommand::ListChecksCommand/2")
            WOAP2_LOG_DEBUG("ListChecksCommand")
            WOAP2_LOG_TRACE(" < ListChecksCommand::ListChecksCommand/2")
        }
    
        virtual ~ListChecksCommand()
        {
            WOAP2_LOG_TRACE(" > ListChecksCommand::~ListChecksCommand")
            WOAP2_LOG_TRACE(" < ListChecksCommand::~ListChecksCommand")
        }
        
        /**
         * Execute list checks.
         * @return list checks resylt string.
         */
        virtual std::string execute()
        {
            WOAP2_LOG_TRACE(" > ListChecksCommand::execute")
            WOAP2_LOG_DEBUG("Executing ListChecksCommand.")
    
            std::pair<int, std::string> checks = checks_factory.list_checks();
    
            WOAP2_LOG_TRACE(" < ListChecksCommand::execute")
            
            return "#keycount=" + std::to_string(checks.first * 3 + 3) +
                   "#responseaction=listchecks#sessionid=1#checkcount=" +
                   std::to_string(checks.first) + checks.second;
        }
    };
    
    /**
     * Get check metadata (define check) command.
     */
    class DefineCheckCommand : public CommandBase
    {
    protected:
        /**
         * check ID
         */
        std::string check_id;

    public:
        /**
         * Constructor. Extracts check name for futher usage.
         * @param params command parameters map.
         * @param _checks_factory existing checks factory.
         */
        DefineCheckCommand(ParamsMap const &params, checks::CheckFactory const &_checks_factory)
                : CommandBase(_checks_factory)
        {
            WOAP2_LOG_TRACE(" > DefineCheckCommand::DefineCheckCommand/2")
    
            ParamsMap::const_iterator check_it = params.find("checkid");
    
            if(check_it != params.end())
            {
                check_id = check_it->second;
            }
    
            WOAP2_LOG_DEBUG(std::string("DefineCheckCommand with checkid=") + check_id)
            WOAP2_LOG_TRACE(" < DefineCheckCommand::DefineCheckCommand/2")
        }
    
        virtual ~DefineCheckCommand()
        {
            WOAP2_LOG_TRACE(" > DefineCheckCommand::~DefineCheckCommand")
            WOAP2_LOG_TRACE(" < DefineCheckCommand::~DefineCheckCommand")
        }
        
        /**
         * Execute command.
         * @return result string.
         */
        virtual std::string execute()
        {
            WOAP2_LOG_TRACE(" > DefineCheckCommand::execute")
            WOAP2_LOG_DEBUG(std::string("Executing DefineCheckCommand. checkid=") + check_id)
            
            if(check_id.length() == 0)
            {
                WOAP2_LOG_WARN("DefineCheckCommand::execute: empty check_id")
                WOAP2_LOG_TRACE(" < DefineCheckCommand::execute")
    
                return "#keycount=4#responseaction=definecheck#sessionid=1#checkid=" +
                       check_id + "#errormessage=Check not found#";
            }
    
            std::unique_ptr <checks::CheckBase> check = checks_factory.get_check(check_id, true);
            std::string params_definition = check->get_params_definition();
    
            WOAP2_LOG_TRACE(" < DefineCheckCommand::execute")
    
            return "#keycount=4#responseaction=definecheck#sessionid=1#checkid=" +
                   check_id + "#parameters=\"" + params_definition + "\"#";
        }
    };
    
    /**
     * Get Combo command actually does the same thing as Define Check but with different responseaction
     */
    class GetComboCommand : public DefineCheckCommand
    {
    public:
        GetComboCommand(ParamsMap const &params, checks::CheckFactory const &_checks_factory)
                : DefineCheckCommand(params, _checks_factory)
        {
            WOAP2_LOG_TRACE(" > GetComboCommand::GetComboCommand/2")
            WOAP2_LOG_DEBUG(std::string("GetComboCommand with checkid=") + check_id)
            WOAP2_LOG_TRACE(" < GetComboCommand::GetComboCommand/2")
        }
    
        virtual ~GetComboCommand()
        {
            WOAP2_LOG_TRACE(" < GetComboCommand::~GetComboCommand")
            WOAP2_LOG_TRACE(" > GetComboCommand::~GetComboCommand")
        }
        
        /**
         * Execute command.
         * @return result string.
         */
        virtual std::string execute()
        {
            WOAP2_LOG_TRACE(" > GetComboCommand::execute")
            WOAP2_LOG_DEBUG(std::string("Executing GetComboCommand. checkid=") + check_id)
            
            if(check_id.length() == 0)
            {
                WOAP2_LOG_WARN("GetComboCommand::execute: empty check_id")
                WOAP2_LOG_TRACE(" < GetComboCommand::execute")
    
                return "#keycount=4#responseaction=getcomboparamvalues#sessionid=1#checkid=" +
                       check_id + "#errormessage=Check not found#";
            }
    
            std::unique_ptr <checks::CheckBase> check = checks_factory.get_check(check_id, true);
            std::string params_definition = check->get_params_definition();
    
            WOAP2_LOG_TRACE(" < GetComboCommand::execute")
            
            return "#keycount=4#responseaction=getcomboparamvalues#sessionid=1#checkid=" +
                   check_id + "#parameters=\"" + params_definition + "\"#";
        }
        
    };
    
    /***
     * Login command.
     */
    class LoginCommand : public CommandBase
    {
    public:
        LoginCommand(checks::CheckFactory const &_checks_factory)
                : CommandBase(_checks_factory)
        {
            WOAP2_LOG_TRACE(" > LoginCommand::LoginCommand/1")
            WOAP2_LOG_DEBUG("LoginCommand")
            WOAP2_LOG_TRACE(" < LoginCommand::LoginCommand/1")
        }
        
        virtual ~LoginCommand()
        {
            WOAP2_LOG_TRACE(" > LoginCommand::~LoginCommand")
            WOAP2_LOG_TRACE(" < LoginCommand::~LoginCommand")
        }
        
        virtual std::string execute() override
        {
            WOAP2_LOG_TRACE(" > LoginCommand::execute")
            WOAP2_LOG_DEBUG("Executing LoginCommand.")
            WOAP2_LOG_TRACE(" < LoginCommand::execute")
            
            return "#keycount=3#responseaction=login#sessionid=1#registrationlevel=standard#";
        }
    };
    
    /***
     * Logout command
     */
    class LogoutCommand : public CommandBase
    {
    public:
        LogoutCommand(ParamsMap const &params, checks::CheckFactory const &_checks_factory)
                : CommandBase(_checks_factory)
        {
            WOAP2_LOG_TRACE(" > LogoutCommand::LogoutCommand/1")
            WOAP2_LOG_DEBUG("LogoutCommand")
            WOAP2_LOG_TRACE(" < LogoutCommand::LogoutCommand/1")
        }
        
        virtual ~LogoutCommand()
        {
            WOAP2_LOG_TRACE(" > LogoutCommand::~LogoutCommand")
            WOAP2_LOG_TRACE(" < LogoutCommand::~LogoutCommand")
        }
        
        virtual std::string execute() override
        {
            WOAP2_LOG_TRACE(" > LogoutCommand::execute")
            WOAP2_LOG_DEBUG("Executing LogoutCommand.")
            WOAP2_LOG_TRACE(" < LogoutCommand::execute")
            return "";
        }
    };
    
    /**
     * Commands factory
     */
    class CommandsFactory
    {
        template<class T>
        static std::unique_ptr<CommandBase>
        creator_func(ParamsMap const &params, checks::CheckFactory const &checks_factory)
        {
            return std::unique_ptr<CommandBase>(new T(params, checks_factory));
        }
        
        /**
         * Command creator function type.
         */
        typedef std::unique_ptr<CommandBase>(*CommandCreatorFunc)(ParamsMap const &, checks::CheckFactory const &);
    
        /**
         * Commands creators map type.
         */
        typedef std::map<std::string, CommandCreatorFunc> CommandsMap;
    
        /**
         * Commands creators map.
         */
        CommandsMap commands_map;
    
        /**
         * Checks factory.
         */
        checks::CheckFactory check_factory;
    
    public:
    
        /**
         * Constructor. Fills the commands map.
         */
        CommandsFactory(Settings &settings)
                : check_factory(settings)
        {
            WOAP2_LOG_TRACE(" > CommandsFactory::CommandsFactory")
            
            commands_map["listchecks"] = creator_func<ListChecksCommand>;
            commands_map["definecheck"] = creator_func<DefineCheckCommand>;
            commands_map["runcheck"] = creator_func<RunCheckCommand>;
            commands_map["getcomboparamvalues"] = creator_func<GetComboCommand>;
            commands_map["logout"] = creator_func<LogoutCommand>;
    
            WOAP2_LOG_TRACE(" < CommandsFactory::CommandsFactory")
        }
    
        /**
         * Get command object by command string.
         * @param input command string.
         * @return command object.
         */
        std::unique_ptr<CommandBase> get_command(std::string const &input) const
        {
            WOAP2_LOG_TRACE(" > CommandsFactory::get_command")
            
            ParamsMap params_map = utils::get_params_map(input);
            
            ParamsMap::const_iterator param_it = params_map.find("keycount");
            
            if(param_it == params_map.end())
            {
                WOAP2_LOG_ERROR("Protocol Error: no keycount.")
                WOAP2_LOG_TRACE(" < CommandsFactory::get_command")
                return std::unique_ptr<CommandBase>(new ErrorCommand("no keycount"));
            }
            
            param_it = params_map.find("sessionid");
            
            if(param_it == params_map.end())
            {
                WOAP2_LOG_ERROR("Protocol Error: no sessionid.")
                WOAP2_LOG_TRACE(" < CommandsFactory::get_command")
                return std::unique_ptr<CommandBase>(new ErrorCommand("no sessionid"));
            }
            
            param_it = params_map.find("requestaction");
            
            if(param_it == params_map.end())
            {
                WOAP2_LOG_TRACE(" < CommandsFactory::get_command")
                WOAP2_LOG_ERROR("Protocol Error: no requestaction.")
                return std::unique_ptr<CommandBase>(new ErrorCommand("no requestaction"));
            }
            
            const std::string &action = param_it->second;
    
            WOAP2_LOG_DEBUG(std::string("Getting command for action ") + action)
            
            CommandsMap::const_iterator cmd_it = commands_map.find(action);
            
            if(cmd_it == commands_map.end())
            {
                WOAP2_LOG_ERROR(std::string("Unknown action ") + action)
                WOAP2_LOG_TRACE(" < CommandsFactory::get_command")
                return std::unique_ptr<CommandBase>(new UnknownCommand(action));
            }
    
            WOAP2_LOG_TRACE(" < CommandsFactory::get_command")
            return cmd_it->second(params_map, check_factory);
        }
    
        std::unique_ptr<CommandBase> get_login_command()
        {
            WOAP2_LOG_TRACE(" > CommandsFactory::get_login_command")
            WOAP2_LOG_TRACE(" < CommandsFactory::get_login_command")
            return std::unique_ptr<CommandBase>(new LoginCommand(check_factory));
        }
    };
    
}
#endif //WOAP2_COMMANDS_H
