//
// Created by kirdyk on 30.11.16.
//

#ifndef WOAP2_HTTPCLIENT_H
#define WOAP2_HTTPCLIENT_H

#include <string>
#include <iostream>
#include <boost/bind.hpp>
#include <boost/asio.hpp>
#include <boost/asio/ssl.hpp>
#include "utils.h"
#include "logger.h"

namespace http
{
    using boost::asio::ip::tcp;
    
    /***
     * base class for HTTP(S) clients
     */
    class HttpClientBase
    {
    private:
        /***
         * failed flad
         */
        bool failed;
        /***
         * error message
         */
        std::string message;
        /***
         * HTTP status code
         */
        unsigned int status_code;
        /***
         * HTTP status message
         */
        std::string status_message;
        // URL parts
        std::string host, service, path;
    
        // text body found and text found flags
        bool text_body, text_found;
    
        /***
         * Text to search
         */
        std::string text_to_search;
    
    protected:
    
        /***
         * Constructor
         * @param host
         * @param service
         * @param path
         */
        HttpClientBase(
                const std::string &host,
                const std::string &service,
                const std::string &path,
                const std::string &search
        ) :
                failed(false),
                message(""),
                status_code(0),
                status_message("__UNKNOWN__"),
                host(host),
                service(service),
                path(path),
                text_body(false),
                text_found(false),
                text_to_search(search)
        {
            
        }
    
        /***
         * Read headers and check if content-type is text.
         * @param response http response buffer
         */
        void read_check_text_content(boost::asio::streambuf &response)
        {
            std::istream response_stream(&response);
        
            // Read headers out of stream
            std::string line;
            while(std::getline(response_stream, line) && line != "\r")
            {
                std::transform(line.begin(), line.end(), line.begin(), ::tolower);
                if(line.find("content-type:") == 0)
                {
                    if(line.find("text") != std::string::npos)
                    {
                        text_body = true;
                    }
                }
            }
        }
    
        /***
         * Search for the text
         * @param response http response buffer
         */
        void find_text(boost::asio::streambuf &response)
        {
            text_found = false;
        
            if(text_to_search == "")
            {
                text_found = true;
                return;
            }
        
            if(text_body)
            {
                std::istream response_stream(&response);
            
                std::string line;
            
                while(std::getline(response_stream, line))
                {
                    if(line.find(text_to_search) != std::string::npos)
                    {
                        text_found = true;
                    }
                }
            }
        }
        
    public:
    
        /***
         * Destructor
         */
        virtual ~HttpClientBase()
        {
        
        }
    
        /***
         * failed flag getter
         * @return failed flag
         */
        bool is_failed() const
        {
            return failed;
        }
    
        /***
         * failed flag setter
         * @param val
         */
        void set_failed(const bool val)
        {
            failed = val;
        }
    
        /***
         * error message getter
         * @return error message
         */
        std::string get_message() const
        {
            return message;
        }
    
        /***
         * error message setter
         * @param val
         */
        void set_message(const std::string &val)
        {
            message = val;
        }
    
        /***
         * status code getter
         * @return HTTP status code
         */
        unsigned int get_status() const
        {
            return status_code;
        }
    
        /***
         * status code setter
         * @param val
         */
        void set_status(const unsigned int val)
        {
            status_code = val;
        }
    
        /***
         * status message getter
         * @return HTTP status message
         */
        std::string get_status_message() const
        {
            return utils::trim(status_message);
        }
    
        /***
         * status message setter
         * @param val
         */
        void set_status_message(const std::string &val)
        {
            status_message = val;
        }
    
        /***
         * is reponse content text
         */
        bool is_text_response() const
        {
            return text_body;
        }
    
        /***
         * Do we need to perform a search
         */
        bool need_to_search() const
        {
            return text_body && text_to_search.length() > 0;
        }
    
        /***
         * have we found a text
         */
        bool is_text_found() const
        {
            return text_found;
        }
    };
    
    /***
     * boost::asio HTTPS client
     */
    class HttpsClient : public HttpClientBase
    {
    private:
        /***
         * name resolver
         */
        tcp::resolver resolver;
        /***
         * SSL context
         */
        boost::asio::ssl::context ctx;
        /***
         * SSL enabled socket
         */
        boost::asio::ssl::stream<boost::asio::ip::tcp::socket> socket;
        /***
         * request buffer
         */
        boost::asio::streambuf request;
        /***
         * response buffer
         */
        boost::asio::streambuf response;
    
    public:
        /***
         * Constructor
         * @param io_service boost::asio IO service instance
         * @param host
         * @param service
         * @param path
         */
        HttpsClient(
                boost::asio::io_service &io_service,
                const std::string &host,
                const std::string &service,
                const std::string &path,
                const std::string &search
        ) :
                HttpClientBase(host, service, path, search),
                resolver(io_service),
                ctx(boost::asio::ssl::context::sslv23),
                socket(io_service, ctx)
        {
            std::ostream request_stream(&request);
            // construct request
            request_stream << "GET " << path << " HTTP/1.0\r\n"
                           << "Host: " << host << "\r\n"
                           << "Accept: */*\r\n"
                           << "Connection: close\r\n"
                           << "\r\n";
    
            // resolve host name. the callback handle_resolve will be called
            tcp::resolver::query query(host, service);
            resolver.async_resolve(query,
                                   boost::bind(
                                           &HttpsClient::handle_resolve,
                                           this,
                                           boost::asio::placeholders::error,
                                           boost::asio::placeholders::iterator
                                   ));
        }
    
        /***
         * Destructor
         */
        virtual ~HttpsClient()
        {
        
        }
        
    private:
        /***
         * Dummy certificate verification callback. Trusts everything.
         */
        bool verify_certificate(bool preverified,
                                boost::asio::ssl::verify_context &ctx)
        {
            return true;
        }
    
        /***
         * Name resolved callback.
         * @param error
         * @param endpoint_iterator resolve result
         */
        void handle_resolve(
                const boost::system::error_code &error,
                tcp::resolver::iterator endpoint_iterator)
        {
            // if host was resolved. start connection. callback will be called afterwards.
            if(!error)
            {
                socket.set_verify_mode(boost::asio::ssl::verify_peer);
                socket.set_verify_callback(
                        boost::bind(&HttpsClient::verify_certificate, this, _1, _2));
                
                boost::asio::async_connect(
                        socket.lowest_layer(), endpoint_iterator,
                        boost::bind(
                                &HttpsClient::handle_connect,
                                this,
                                boost::asio::placeholders::error
                        )
                );
            }
            else
            {
                WOAP2_LOG_ERROR(std::string("Resolve failed: ") + error.message());
                set_failed(true);
                set_message(error.message());
            }
        }
    
        /***
         * connect event callback.
         * @param error
         */
        void handle_connect(const boost::system::error_code &error)
        {
            // if connected -- do SSL handshake and move to next callback.
            if(!error)
            {
                socket.async_handshake(boost::asio::ssl::stream_base::client,
                                       boost::bind(&HttpsClient::handle_handshake, this,
                                                   boost::asio::placeholders::error));
            }
            else
            {
                WOAP2_LOG_ERROR(std::string("Connect failed: ") + error.message());
                set_failed(true);
                set_message(error.message());
            }
        }
    
        /***
         * SSL handshake event callback
         * @param error
         */
        void handle_handshake(const boost::system::error_code &error)
        {
            // if SSL is established -- send request and move to next callback.
            if(!error)
            {
                boost::asio::async_write(
                        socket, request,
                        boost::bind(
                                &HttpsClient::handle_write_request,
                                this,
                                boost::asio::placeholders::error
                        )
                );
            }
            else
            {
                WOAP2_LOG_ERROR(std::string("Handshake failed: ") + error.message());
                set_failed(true);
                set_message(error.message());
            }
        }
    
        /***
         * Send data callback.
         * @param error
         */
        void handle_write_request(const boost::system::error_code &error)
        {
            // If request was sent. Read the response until "\r\n" and call next method.
            if(!error)
            {
                boost::asio::async_read_until(
                        socket, response, "\r\n",
                        boost::bind(
                                &HttpsClient::handle_read_status,
                                this,
                                boost::asio::placeholders::error
                        )
                );
            }
            else
            {
                WOAP2_LOG_ERROR(std::string("Write request failed: ") + error.message());
                set_failed(true);
                set_message(error.message());
            }
        }
    
        /***
         * status read event callback
         * @param error
         */
        void handle_read_status(const boost::system::error_code &error)
        {
            // If status was read. Parse it.
            // "HTTP/1.1 200 OK\r\n"
            if(!error)
            {
                std::istream response_stream(&response);
                
                std::string http_version;
                response_stream >> http_version;
                
                unsigned int sc = 0;
                response_stream >> sc;
                set_status(sc);
                
                std::string sm("__UNKNOWN__");
                std::getline(response_stream, sm);
                set_status_message(sm);
    
                if(!response_stream || http_version.substr(0, 5) != "HTTP/")
                {
                    WOAP2_LOG_ERROR(std::string("Invalid response"));
                    set_failed(true);
                    set_message("Invalid response");
                }
    
                // If you want to read the body -- do the following.
                boost::asio::async_read_until(
                        socket, response, "\r\n\r\n",
                        boost::bind(
                                &HttpsClient::handle_read_headers,
                                this,
                                boost::asio::placeholders::error
                        )
                );
            }
            else
            {
                WOAP2_LOG_ERROR(std::string("Read status failed: ") + error.message());
                set_failed(true);
                set_message(error.message());
            }
        }
    
        /***
         * headers read callback
         * @param error
         */
        void handle_read_headers(const boost::system::error_code &error)
        {
            if(!error)
            {
                // read headers and check if content is text
                read_check_text_content(response);
    
                // search if needed.
                if(need_to_search())
                {
                    // Read body.
                    boost::asio::async_read(
                            socket, response,
                            boost::asio::transfer_at_least(1),
                            boost::bind(
                                    &HttpsClient::handle_read_body,
                                    this,
                                    boost::asio::placeholders::error
                            )
                    );
                }
            }
            else
            {
                WOAP2_LOG_ERROR(std::string("Read headers failed: ") + error.message());
                set_failed(true);
                set_message(error.message());
            }
        }
    
        /***
         * response body read callback
         * @param error
         */
        void handle_read_body(const boost::system::error_code &error)
        {
            if(!error || error != boost::asio::error::eof)
            {
                find_text(response);
            }
            else
            {
                WOAP2_LOG_ERROR(std::string("Read body failed: ") + error.message());
                set_failed(true);
                set_message(error.message());
            }
        }
        
    };
    
    /***
     * boost::asio HTTP client
     */
    class HttpClient : public HttpClientBase
    {
    private:
        /***
         * name resolver
         */
        tcp::resolver resolver;
        /***
         * TCP socket
         */
        tcp::socket socket;
        /***
         * request buffer
         */
        boost::asio::streambuf request;
        /***
         * response buffer
         */
        boost::asio::streambuf response;
    
    public:
        /***
         * Constructor
         * @param io_service boost::asio IO service instance
         * @param host
         * @param service
         * @param path
         */
        HttpClient(
                boost::asio::io_service &io_service,
                const std::string &host,
                const std::string &service,
                const std::string &path,
                const std::string &search
        ) :
                HttpClientBase(host, service, path, search),
                resolver(io_service),
                socket(io_service)
        {
            std::ostream request_stream(&request);
            // construct request
            request_stream << "GET " << path << " HTTP/1.0\r\n"
                           << "Host: " << host << "\r\n"
                           << "Accept: */*\r\n"
                           << "Connection: close\r\n"
                           << "\r\n";
    
            // resolve host name. the callback handle_resolve will be called
            tcp::resolver::query query(host, service);
            resolver.async_resolve(query,
                                   boost::bind(
                                           &HttpClient::handle_resolve,
                                           this,
                                           boost::asio::placeholders::error,
                                           boost::asio::placeholders::iterator
                                   ));
        }
    
        virtual ~HttpClient()
        {
        
        }
        
    private:
        /***
         * Name resolved callback.
         * @param error
         * @param endpoint_iterator resolve result
         */
        void handle_resolve(
                const boost::system::error_code &error,
                tcp::resolver::iterator endpoint_iterator)
        {
            // if host was resolved. start connection. callback will be called afterwards.
            if(!error)
            {
                boost::asio::async_connect(
                        socket, endpoint_iterator,
                        boost::bind(
                                &HttpClient::handle_connect,
                                this,
                                boost::asio::placeholders::error
                        )
                );
            }
            else
            {
                WOAP2_LOG_ERROR(std::string("Resolve failed: ") + error.message());
                set_failed(true);
                set_message(error.message());
            }
        }
    
        /***
         * connect event callback.
         * @param error
         */
        void handle_connect(const boost::system::error_code &error)
        {
            // if connected -- send request and move to next callback.
            if(!error)
            {
                boost::asio::async_write(
                        socket, request,
                        boost::bind(
                                &HttpClient::handle_write_request,
                                this,
                                boost::asio::placeholders::error
                        )
                );
            }
            else
            {
                WOAP2_LOG_ERROR(std::string("Connect failed: ") + error.message());
                set_failed(true);
                set_message(error.message());
            }
        }
    
        /***
         * Send data callback.
         * @param error
         */
        void handle_write_request(const boost::system::error_code &error)
        {
            // If request was sent. Read the response until "\r\n" and call next method.
            if(!error)
            {
                boost::asio::async_read_until(
                        socket, response, "\r\n",
                        boost::bind(
                                &HttpClient::handle_read_status,
                                this,
                                boost::asio::placeholders::error
                        )
                );
            }
            else
            {
                WOAP2_LOG_ERROR(std::string("Write request failed: ") + error.message());
                set_failed(true);
                set_message(error.message());
            }
        }
    
        /***
         * status read event callback
         * @param error
         */
        void handle_read_status(const boost::system::error_code &error)
        {
            // If status was read. Parse it.
            // "HTTP/1.1 200 OK\r\n"
            if(!error)
            {
                std::istream response_stream(&response);
                
                std::string http_version;
                response_stream >> http_version;
                
                unsigned int sc = 0;
                response_stream >> sc;
                set_status(sc);
                
                std::string sm("__UNKNOWN__");
                std::getline(response_stream, sm);
                set_status_message(sm);
                
                if(!response_stream || http_version.substr(0, 5) != "HTTP/")
                {
                    WOAP2_LOG_ERROR(std::string("Invalid response"));
                    set_failed(true);
                    set_message("Invalid response");
                }
    
                // Read headers.
                boost::asio::async_read_until(
                        socket, response, "\r\n\r\n",
                        boost::bind(
                                &HttpClient::handle_read_headers,
                                this,
                                boost::asio::placeholders::error
                        )
                );
            }
            else
            {
                WOAP2_LOG_ERROR(std::string("Read status failed: ") + error.message());
                set_failed(true);
                set_message(error.message());
            }
        }
    
        /***
         * Read headers callback
         * @param error
         */
        void handle_read_headers(const boost::system::error_code &error)
        {
            if(!error)
            {
                read_check_text_content(response);
            
                if(need_to_search())
                {
                    // Read body.
                    boost::asio::async_read(
                            socket, response,
                            boost::asio::transfer_at_least(1),
                            boost::bind(
                                    &HttpClient::handle_read_body,
                                    this,
                                    boost::asio::placeholders::error
                            )
                    );
                }
            }
            else
            {
                WOAP2_LOG_ERROR(std::string("Read headers failed: ") + error.message());
                set_failed(true);
                set_message(error.message());
            }
        }
    
        /***
         * response body read callback
         * @param error
         */
        void handle_read_body(const boost::system::error_code &error)
        {
            if(!error || error != boost::asio::error::eof)
            {
                find_text(response);
            }
            else
            {
                WOAP2_LOG_ERROR(std::string("Read body failed: ") + error.message());
                set_failed(true);
                set_message(error.message());
            }
        }
    };
    
}
#endif //WOAP2_HTTPCLIENT_H
