#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>



#include "programoption.h"
#include "woapworker.h"

#include "logger.h"

#include "version.h"

/**
 * Daemonize process.
 * @param name process name.
 */
static void daemonize(const char *name, DaemonSettings &settings)
{
    WOAP2_LOG_TRACE(" > daemonize")
    pid_t pid;
    
    /* Fork off the parent process */
    pid = fork();
    
    /* An error occurred */
    if(pid < 0)
    {
        int err = errno;
        WOAP2_LOG_ERROR(std::string("First fork failed: ") + std::string(strerror(err)));
        exit(EXIT_FAILURE);
    }
    
    /* Success: Let the parent terminate */
    if(pid > 0)
    {
        exit(EXIT_SUCCESS);
    }
    
    /* On success: The child process becomes session leader */
    if(setsid() < 0)
    {
        int err = errno;
        WOAP2_LOG_ERROR(std::string("setsid failed: ") + std::string(strerror(err)));
        exit(EXIT_FAILURE);
    }
    
    /* Catch, ignore and handle signals */
    //TODO: Implement a working signal handler */
    signal(SIGCHLD, SIG_IGN);
    signal(SIGHUP, SIG_IGN);
    
    /* Fork off for the second time*/
    pid = fork();
    
    /* An error occurred */
    if(pid < 0)
    {
        int err = errno;
        WOAP2_LOG_ERROR(std::string("Second fork failed: ") + std::string(strerror(err)));
        exit(EXIT_FAILURE);
    }
    
    /* Success: Let the parent terminate */
    if(pid > 0)
    {
        exit(EXIT_SUCCESS);
    }
    
    /* Set new file permissions */
    umask(0);
    
    /* Change the working directory to the root directory */
    /* or another appropriated directory */
    chdir("/");
    
    /* Close all open file descriptors */
    long x;
    for(x = sysconf(_SC_OPEN_MAX); x > 0; x--)
    {
        close(x);
    }
    
    /* Open the log file */
    
    int nullfd;
    nullfd = open("/dev/null", O_RDWR, 0);
    
    dup2(nullfd, STDIN_FILENO);
    dup2(nullfd, STDOUT_FILENO);
    dup2(nullfd, STDERR_FILENO);
    
    if(settings.get_log_file_path().length() > 0)
    {
        WOAP2_LOG_DEBUG(std::string("Opening ") + settings.get_log_file_path() + " for logging.");
        int logfd = open(settings.get_log_file_path().c_str(), O_APPEND | O_CREAT | O_WRONLY, 0755);
        
        if(logfd > 0)
        {
            dup2(logfd, STDERR_FILENO);
            
            if(logfd > STDERR_FILENO)
            {
                close(logfd);
            }
        }
    }
    
    if(nullfd > STDERR_FILENO)
    {
        close(nullfd);
    }
    
    WOAP2_LOG_TRACE(" < daemonize")
}

WoapWorker *_worker;

void sigint_handler(int p)
{
    WOAP2_LOG_TRACE(" > sigint_handler")
    WOAP2_LOG_DEBUG(std::string("Signal") + std::to_string(p) + " caught")
    
    if(_worker)
    {
        _worker->stop();
        _worker = nullptr;
    }
    
    WOAP2_LOG_TRACE(" < sigint_handler")
}

/**
 * Entry point.
 * @param argc
 * @param argv
 * @return
 */
int main(int argc, char **argv)
{
    init_snmp("woap2");
    
    logger::Logger::current_logger->set_log_level(logger::Logger::WARN);
    
    std::string ini_path = "/etc/woap2/woap.ini";
    
    if(argc > 1)
    {
        struct stat df_stat;
        
        if(!stat(argv[1], &df_stat))
        {
            ini_path = argv[1];
        }
    }
    
    Settings settings(ini_path);
    DaemonSettings daemon_settings = settings.get_daemon_settings();
    
    logger::Logger::current_logger->set_log_level(daemon_settings.get_log_level());
    
    if(daemon_settings.get_log_file_path().length() > 0)
    {
        int logfd = open(daemon_settings.get_log_file_path().c_str(), O_APPEND | O_CREAT | O_WRONLY, 0755);
        
        if(logfd < 0)
        {
            int err = errno;
            WOAP2_LOG_ERROR(std::string("Can not open log file: ") + strerror(err))
            exit(EXIT_FAILURE);
        }
        else
        {
            dup2(logfd, STDERR_FILENO);
        }
        
        close(logfd);
    }
    
    WOAP2_LOG_TRACE("Logging set up.")
    WOAP2_LOG_INFO("Starting woap2-"
                     Proj_VERSION_STR)
    
    if(daemon_settings.is_daemonize())
    {
        WOAP2_LOG_DEBUG("Daemonization requested.")
        daemonize(argv[0], daemon_settings);
        WOAP2_LOG_DEBUG("Daemonized.")
    }
    
    struct sigaction action;
    action.sa_handler = sigint_handler;
    sigemptyset(&action.sa_mask);
    action.sa_flags = 0;
    
    sigaction(SIGINT, &action, NULL);
    sigaction(SIGTERM, &action, NULL);
    
    WoapWorker worker(settings);
    _worker = &worker;
    
    worker.do_work();
    
    WOAP2_LOG_INFO("Exiting.")
    
    return 0;
}

