#include "programoption.h"


Settings::Settings(const std::string &ini_file_path)
{
    password_map = std::shared_ptr<password_map_type>(new password_map_type());
    firewall_map = std::shared_ptr<firewall_map_type>(new firewall_map_type());
    checks_set = std::shared_ptr<allowed_checks_type>(new allowed_checks_type());
    
    boost::regex dash_re("\\s*-\\s*");
    boost::regex comma_re("\\s*,\\s*");
    
    try
    {
        boost::property_tree::ptree prop_tree;
        
        boost::property_tree::ini_parser::read_ini(ini_file_path, prop_tree);
        
        int port = prop_tree.get("agent.port", DEF_PORT);
        std::string log_file = prop_tree.get("agent.log", DEF_LOG);
        bool daemon = prop_tree.get("agent.daemonize", DEF_DAEMONIZE);
        std::string keyFolder = prop_tree.get("agent.keyFolder", DEF_KEYFOLDER);
        std::string dsaKey = prop_tree.get("agent.dsaKey", DEF_DSA);
        std::string rsaKey = prop_tree.get("agent.rsaKey", DEF_RSA);
        std::string logLvlStr = prop_tree.get("agent.loglevel", DEF_LOG_LEVEL_STR);
        std::string checksStr = prop_tree.get("agent.checks", "");
    
        std::vector<std::string> checks_vec = utils::split(checksStr, comma_re);
    
        for(std::vector<std::string>::const_iterator it = checks_vec.cbegin(); it != checks_vec.cend(); ++it)
        {
            if(checks_set->find(*it) == checks_set->end())
            {
                checks_set->insert(*it);
            }
        }
        
        logger::Logger::LEVEL logLevel = logger::Logger::parse_level(logLvlStr);
        
        
        daemon_settings = DaemonSettings(
                port, log_file, logLevel,
                keyFolder, dsaKey, rsaKey,
                daemon);
        
        ptree password_tree = prop_tree.get_child("password");
        
        for(ptree::const_iterator it = password_tree.begin(); it != password_tree.end(); ++it)
        {
            (*password_map)[(it->first)] = (it->second).data();
        }
        
        ptree::assoc_iterator fw = prop_tree.find("firewall");
        
        if(fw != prop_tree.not_found())
        {
            ptree firewall_tree = prop_tree.get_child("firewall");
            for(ptree::const_iterator it = firewall_tree.begin(); it != firewall_tree.end(); ++it)
            {
                std::string data = (it->second).data();
                
                std::vector<std::string> data_parts = utils::split(data, dash_re);
                
                if(data_parts.size() != 2)
                {
                    WOAP2_LOG_ERROR(std::string("Incorrect firewall record for user ") + it->first)
                    continue;
                }
                else
                {
                    std::string range_start = data_parts[0];
                    std::string range_end = data_parts[1];
                    
                    bool start_colon = (range_start.find(':') != std::string::npos);
                    bool end_colon = (range_end.find(':') != std::string::npos);
                    
                    if(start_colon != end_colon)
                    {
                        WOAP2_LOG_ERROR(std::string("Incorrect firewall record for user ") + it->first)
                        continue;
                    }
                    else
                    {
                        (*firewall_map)[(it->first)] = std::make_pair(range_start, range_end);
                    }
                }
            }
        }
        
    }
    catch(boost::property_tree::ptree_error &ex)
    {
        WOAP2_LOG_ERROR(std::string("Error parsing ini file: ") + ex.what());
        
        exit(-1);
    }
}
