#ifndef PROGRAMOPTION_H
#define PROGRAMOPTION_H

#include <getopt.h>
#include <boost/lexical_cast.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/ini_parser.hpp>
#include <vector>
#include <map>
#include <set>

#include "logger.h"
#include "utils.h"

#define DEF_PORT 20012
#define DEF_LOG ""
#define DEF_KEYFOLDER ""
#define DEF_DSA "ssh_host_dsa_key"
#define DEF_RSA "ssh_host_rsa_key"
#define DEF_DAEMONIZE false
#define DEF_LOG_LEVEL_STR "Error"
#define DEF_LOG_LEVEL logger::Logger::ERROR

//template <class T> class ProgramOption
//{
//    std::string long_name;
//    char short_name;
//
//public:
//    ProgramOption(const std::string &long_name, char short_name)
//        : long_name(long_name), short_name(short_name)
//    {
//
//    }
//
//    std::string get_long_name() const
//    {
//        return long_name;
//    }
//
//    char get_short_name() const
//    {
//        return short_name;
//    }
//
//    T parse(const std::string &value)
//    {
//        return boost::lexical_cast<T>(value);
//    }
//};


//class ProgramOptions
//{
//
//
//};


/**
 * Daemon settings.
 */
class DaemonSettings
{
    /**
     * listen port
     */
    int port;
    /**
     * path to log file.
     */
    std::string logFile;
    /**
     * log level
     */
    logger::Logger::LEVEL log_level;
    
    /**
     * daemonize flag
     */
    bool daemonize;
    
    /**
     * ssh keys folder.
     */
    std::string keys_folder;
    /**
     * ssh DSA key file name.
     */
    std::string dsa_key;
    /**
     * ssh RSA key file name.
     */
    std::string rsa_key;


public:
    
    /**
     * Constructor.Initialize default values.
     */
    DaemonSettings()
    {
        port        = DEF_PORT;
        logFile     = DEF_LOG;
        daemonize   = DEF_DAEMONIZE;
        keys_folder = DEF_KEYFOLDER;
        dsa_key     = DEF_DSA;
        rsa_key     = DEF_RSA;
        log_level = DEF_LOG_LEVEL;
    };
    
    
    /**
     * Parameterized constructor
     * @param port
     * @param logFile
     * @param keysFolder
     * @param dsaKey
     * @param rsaKey
     * @param daemonize
     * @return
     */
    DaemonSettings(
            const int port,
            const std::string &logFile,
            const logger::Logger::LEVEL logLevel,
            const std::string &keysFolder,
            const std::string &dsaKey,
            const std::string &rsaKey,
            const bool daemonize = DEF_DAEMONIZE)
        : port(port),
          logFile(logFile),
          log_level(logLevel),
          daemonize(daemonize),
          keys_folder(keysFolder),
          dsa_key(dsaKey),
          rsa_key(rsaKey)
    {
    }
    
    /**
     * Get listen port.
     * @return listen port number.
     */
    int get_port() const
    {
        return port;
    }
    
    /**
     * Get log file path.
     * @return log file path.
     */
    const std::string& get_log_file_path() const
    {
        return logFile;
    }
    
    /**
     * If daemonization needed.
     * @return daemonize flag.
     */
    bool is_daemonize() const
    {
        return daemonize;
    }
    
    /**
     * Get SSH keys folder.
     * @return SSH keys folder.
     */
    const std::string& get_keys_folder() const
    {
        return keys_folder;
    }
    
    /**
     * Get SSH RSA key file name.
     * @return SSH RSA key file name.
     */
    const std::string& get_rsa_key() const
    {
        return rsa_key;
    }
    
    /**
     * Get SSH DSA key file name.
     * @return SSH DSA key file name.
     */
    const std::string& get_dsa_key() const
    {
        return dsa_key;
    }
    
    logger::Logger::LEVEL get_log_level() const
    {
        return log_level;
    }
    
    void set_log_level(logger::Logger::LEVEL log_level)
    {
        DaemonSettings::log_level = log_level;
    }
    
};

/**
 * Program settings class.
 */
class Settings
{
public:
    /**
     * Passwords map type.
     */
    typedef std::map<std::string, std::string> password_map_type;
    typedef std::map<std::string, std::pair<std::string, std::string> > firewall_map_type;
    typedef std::set<std::string> allowed_checks_type;

private:
    /**
     * Property tree type.
     */
    typedef boost::property_tree::ptree ptree;
    
    /**
     * Daemon settings.
     */
    DaemonSettings daemon_settings;
    
    /**
     * Passwords map.
     */
    std::shared_ptr<password_map_type> password_map;
    
    std::shared_ptr<firewall_map_type> firewall_map;
    
    std::shared_ptr<allowed_checks_type> checks_set;
    
public:
    /**
     * Constructor. Reads and parses ini file.
     * @param ini_file_path ini file path.
     */
    Settings(const std::string &ini_file_path);
    
    /**
     * Get daemon settings.
     * @return daemon settings.
     */
    const DaemonSettings& get_daemon_settings() const
    {
        return daemon_settings;
    }
    
    /**
     * Get passwords map.
     * @return passwords map.
     */
    const std::shared_ptr<password_map_type> get_password_map() const
    {
        return password_map;
    }
    
    const std::shared_ptr<firewall_map_type> get_firewall_map() const
    {
        return firewall_map;
    }
    
    const std::shared_ptr<allowed_checks_type> get_checks_set() const
    {
        return checks_set;
    }
};

#endif // PROGRAMOPTION_H
