//
// Created by kirdyk on 06.01.17.
//

#include "snmpclient.h"
#include "utils.h"

namespace snmp
{
    std::mutex SnmpClient::snmp_mutex;
    
    checks::CheckResult SnmpClient::get_value(std::string &oid_name)
    {
        
        //Root session
        struct snmp_session session;
        // Init and fill in properties
        snmp_sess_init(&session);
        
        session.peername = const_cast<char *>(host.c_str());
        session.version = proto_version;
        
        // Set auth credentials according to SNMP version
        if(proto_version == SNMP_VERSION_2c)
        {
            session.community = reinterpret_cast<u_char *>(const_cast<char *>(community.c_str()));
            session.community_len = strlen((char *) session.community);
        }
        
        // Active session
        auto opened_session = std::shared_ptr<struct snmp_session>(snmp_open(&session), snmp_close);
        
        if(!opened_session)
        {
            char *err;
            snmp_error(&session, NULL, NULL, &err);
            throw std::runtime_error(err);
        }
        
        // GET command
        struct snmp_pdu *pdu = snmp_pdu_create(SNMP_MSG_GET);
        
        // OID numerical value storage
        oid anOID[MAX_OID_LEN];
        size_t anOID_len = MAX_OID_LEN;
        
        // Parse an OID string
        {
            std::lock_guard<std::mutex> lock(snmp_mutex);
    
            if(!snmp_parse_oid(oid_name.c_str(), anOID, &anOID_len))
            {
                int xerr;
                xerr = snmp_errno;
                throw std::runtime_error(snmp_api_errstring(xerr));
            }
        }
        
        // Create a command
        snmp_add_null_var(pdu, anOID, anOID_len);
        
        // Execute it synchronously
        int status;
        struct snmp_pdu *raw_response;
        status = snmp_synch_response(opened_session.get(), pdu, &raw_response);
        auto response = std::shared_ptr<struct snmp_pdu>(raw_response, snmp_free_pdu);
        
        // Check status
        if(status == STAT_SUCCESS)
        {
            // Check response
            if(response->errstat == SNMP_ERR_NOERROR)
            {
                checks::CheckResult result = checks::CheckResult{false, false, "", "__NO_VARIABLES__"};
                // If everything is OK - iterate through variables.
                for(variable_list *vars = response->variables; vars; vars = vars->next_variable)
                {
                    // check for errors
                    if(vars->type == SNMP_NOSUCHOBJECT)
                    {
                        throw std::runtime_error("No such object available on this agent at this OID");
                    }
                    else if(vars->type == SNMP_NOSUCHINSTANCE)
                    {
                        throw std::runtime_error("No such instance currently exists at this OID");
                    }
                    else if(vars->type == SNMP_ENDOFMIBVIEW)
                    {
                        throw std::runtime_error(
                                "No more variables left in this MIB View (It is past the end of the MIB tree)");
                    }
                    else
                    {
                        std::string err_string;
                        std::string value_string;
                        bool failed = false;
                        bool test_result = false;
    
                        // Which data type have we got?
                        switch(vars->type)
                        {
                            // This data types are convertinble to long
                            case ASN_INTEGER:
                            case ASN_GAUGE:
                            case ASN_COUNTER:
                            {
                                long value = *(vars->val.integer);
                                long test_value = std::stol(test_value_str);
                                test_result = compare(value, test_value, test_type);
                                value_string = std::to_string(value);
                            }
                                break;
                                // timeticks type is unsigned
                            case ASN_TIMETICKS:
                            {
                                unsigned long value = static_cast<unsigned long>(*(vars->val.integer));
                                unsigned long test_value = std::stoul(test_value_str);
                                test_result = compare(value, test_value, test_type);
                                value_string = std::to_string(value);
                            }
                                break;
                                // 64-bit unsigned stuff
                            case ASN_COUNTER64:
                            case ASN_UNSIGNED64:
                            {
                                char buf[I64CHARSZ + 1];
                                printU64(buf, vars->val.counter64);
                                uint64_t value = (vars->val.counter64->high << 32) | vars->val.counter64->low;
                                uint64_t test_value = std::stoull(test_value_str);
                                test_result = compare(value, test_value, test_type);
                                value_string = buf;
                            }
                                break;
                                // 64-bit signed stuff
                            case ASN_INTEGER64:
                            {
                                char buf[I64CHARSZ + 1];
                                printI64(buf, vars->val.counter64);
                                value_string = buf;
                                int64_t value = (vars->val.counter64->high << 32) | vars->val.counter64->low;
                                int64_t test_value = std::stoll(test_value_str);
                                test_result = compare(value, test_value, test_type);
                            }
                                break;
                                // String
                            case ASN_OCTET_STR:
                            {
                                std::string value = std::string((char *) vars->val.string, vars->val_len);
    
                                size_t char_pos = value.find("#");
    
                                if(char_pos != std::string::npos)
                                {
                                    value.replace(char_pos, 1, "{35}");
                                }
    
                                char_pos = value.find("\"");
    
                                if(char_pos != std::string::npos)
                                {
                                    value.replace(char_pos, 1, "{34}");
                                }
                                
                                test_result = compare(value, test_value_str, test_type);
                                value_string = value;
                            }
                                break;
                                // Null data type. All NULLs are different, that is NULL != NULL.
                            case ASN_NULL:
                            {
                                test_result = false;
                                value_string = "NULL";
                            }
                                break;
                                // Float
                            case ASN_FLOAT:
                            {
                                float value = *(vars->val.floatVal);
                                float test_value = stof(test_value_str);
                                test_result = compare(value, test_value, test_type);
                                value_string = utils::to_string(value);
                            }
                                break;
                                // Double
                            case ASN_DOUBLE:
                            {
                                double value = *(vars->val.doubleVal);
                                double test_value = stod(test_value_str);
                                test_result = compare(value, test_value, test_type);
                                value_string = utils::to_string(value);
                            }
                                break;
//                                case ASN_OBJECT_ID:
//                                {
//                                    char buf[1024];
//                                    snprint_object_identifier(buf, 1024, vars, NULL, NULL, NULL);
//                                    value = buf;
//                                }
//                                    break;
                                // Other data types are not suppoprted.
                            default:
                                err_string = "Unexpected value type: " + std::to_string((int) vars->type);
                                failed = true;
                                break;
                        }
    
                        // Form the result;
                        result = checks::CheckResult
                                {
                                        !failed,
                                        test_result,
                                        !failed
                                        ? oid_name + " = " + value_string
                                        : "",
                                        failed ? err_string : ""
                                };
                    }
                }
                // return it.
                return result;
            }
            else
            {
                // Throw the error
                throw std::runtime_error(snmp_errstring(static_cast<int>(response->errstat)));
            }
        }
        else
        {
            // Throw the error
            char *err;
            snmp_error(opened_session.get(), NULL, NULL, &err);
            throw std::runtime_error(err);
        }
    }
}