//
// Created by kirdyk on 25.10.16.
//

#ifndef WOAP2_SSLCLIENT_H
#define WOAP2_SSLCLIENT_H


#include <cstdlib>
#include <iostream>
#include <boost/bind.hpp>
#include <boost/asio.hpp>
#include <boost/asio/ssl.hpp>

namespace ssl
{
    /***
     * SSL client class
     */
    class client
    {
    private:
        /***
         * SSL-enabled socket
         */
        boost::asio::ssl::stream<boost::asio::ip::tcp::socket> socket_;
    
        /***
         * Certificate age limit
         */
        int limit;
        /***
         * The state of the check, applied to the last certificate in the chain
         */
        bool last_cert_check_state;
        /***
         * Error flag
         */
        bool failed;
        /***
         * message
         */
        std::string message;
    
    public:
        /***
         * Is certificate valid
         */
        bool is_still_valid() const
        {
            return last_cert_check_state;
        }
    
        /***
         * Have we failed
         */
        bool is_failed() const
        {
            return failed;
        }
    
        /***
         * Get the message
         * @return
         */
        std::string get_message() const
        {
            return message;
        }
    
        /***
         * Constructor
         * @param _limit certificate age limit
         * @param io_service
         * @param context
         * @param endpoint_iterator
         */
        client(int _limit,
               boost::asio::io_service &io_service,
               boost::asio::ssl::context &context,
               boost::asio::ip::tcp::resolver::iterator endpoint_iterator)
                : socket_(io_service, context),
                  limit(_limit),
                  last_cert_check_state(false),
                  failed(false)
        {
            // set verifiocation
            socket_.set_verify_mode(boost::asio::ssl::verify_peer);
            socket_.set_verify_callback(
                    boost::bind(&client::verify_certificate, this, _1, _2));
    
            // Connect
            boost::asio::async_connect(socket_.lowest_layer(), endpoint_iterator,
                                       boost::bind(&client::handle_connect, this,
                                                   boost::asio::placeholders::error));
        }
    
        /***
         * Certificate verification routine.
         */
        bool verify_certificate(bool preverified,
                                boost::asio::ssl::verify_context &ctx)
        {
            // The verify callback can be used to check whether the certificate that is
            // being presented is valid for the peer. For example, RFC 2818 describes
            // the steps involved in doing this for HTTPS. Consult the OpenSSL
            // documentation for more details. Note that the callback is called once
            // for each certificate in the certificate chain, starting from the root
            // certificate authority.
            
            failed = false;
    
            // get certificate name
            char subject_name[256];
            X509 *cert = X509_STORE_CTX_get_current_cert(ctx.native_handle());
            X509_NAME_oneline(X509_get_subject_name(cert), subject_name, 256);
            WOAP2_LOG_TRACE(std::string("Verifying ") + subject_name);
    
            // get expiration date
            ASN1_TIME *expDate = X509_get_notAfter(cert);
    
            // Expiration date can be presented in 2 ASN.1 formats.
            // Handle them
            int t = expDate->type;
            struct tm _tm;
            
            if(t == V_ASN1_UTCTIME)
            {
                strptime((const char *) expDate->data, "%y%m%d%H%M%S", &_tm);
            }
            else if(t == V_ASN1_GENERALIZEDTIME)
            {
                strptime((const char *) expDate->data, "%Y%m%d%H%M%S", &_tm);
            }
            else
            {
                last_cert_check_state = false;
                failed = true;
                message = "Unexpected ASN.1 time type: " + std::to_string(t);
                return true;
            }
    
            // Current datetime
            time_t exp_time = mktime(&_tm);
            time_t now_time = time(NULL);
    
            // Calculate difference and convert it do days.
            double diff = difftime(exp_time, now_time);
            int diff_days = (int) (diff / 3600.0 / 24.0);
    
            // print exp date
            char timeStr[64];
            strftime(timeStr, 63, "%F %T", &_tm);
    
            // Check how many days left.
            if(diff_days < 0)
            {
                last_cert_check_state = false;
                message = std::string("Certificate for ") + subject_name + " is expired " + timeStr;
            }
            else if(diff_days >= limit)
            {
                last_cert_check_state = true;
                message = std::string("Certificate for ") + subject_name + " is valid to " + timeStr;
            }
            else
            {
                last_cert_check_state = false;
                message = std::string("Certificate for ") + subject_name +
                          " will expire in " + std::to_string(diff_days) + "days."
                                  " Valid to " + timeStr;
            }
            
            return true;
        }
    
        /***
         * Connected callback
         * @param error
         */
        void handle_connect(const boost::system::error_code &error)
        {
            if(!error)
            {
                // Do SSL nadshake.
                socket_.async_handshake(boost::asio::ssl::stream_base::client,
                                        boost::bind(&client::handle_handshake, this,
                                                    boost::asio::placeholders::error));
            }
            else
            {
                WOAP2_LOG_ERROR(std::string("Connect failed: ") + error.message());
                failed = true;
                message = error.message();
            }
        }
    
        /***
         * Handshake completed callback
         * @param error
         */
        void handle_handshake(const boost::system::error_code &error)
        {
            if(!error)
            {
//                std::cout << "Enter message: ";
//                std::cin.getline(request_, max_length);
//                size_t request_length = strlen(request_);
//
//                boost::asio::async_write(socket_,
//                                         boost::asio::buffer(request_, request_length),
//                                         boost::bind(&client::handle_write, this,
//                                                     boost::asio::placeholders::error,
//                                                     boost::asio::placeholders::bytes_transferred));
            }
            else
            {
                WOAP2_LOG_ERROR(std::string("Handshake failed: ") + error.message());
                failed = true;
                message = error.message();
            }
        }
    };
}

#endif //WOAP2_SSLCLIENT_H
