#include "woapworker.h"
#include "utils.h"
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "logger.h"

#include <string>

#include "version.h"

WoapWorker::WoapWorker(Settings &settings)
        : commands_factory(settings)
{
    WOAP2_LOG_TRACE(" > WoapWorker::WoapWorker")
    
    port = settings.get_daemon_settings().get_port();
    rsa_key = utils::path_combine(
            settings.get_daemon_settings().get_keys_folder(),
            settings.get_daemon_settings().get_rsa_key());
    dsa_key = utils::path_combine(
            settings.get_daemon_settings().get_keys_folder(),
            settings.get_daemon_settings().get_dsa_key());
    
    password_map = settings.get_password_map();
    firewall_map = settings.get_firewall_map();
    
    WOAP2_LOG_TRACE(" < WoapWorker::WoapWorker")
}

bool WoapWorker::authenticate(const std::string &user, const std::string &password, const std::string &client_ip)
{
    WOAP2_LOG_TRACE(" > WoapWorker::authenticate");
    WOAP2_LOG_DEBUG(std::string("Authenticating user ") + user + " with password " + password + " from " + client_ip);
    
    Settings::password_map_type::iterator it = password_map->find(user);
    
    bool result = false;
    
    if(it != password_map->end())
    {
        result = (it->second == password);
    }
    
    WOAP2_LOG_DEBUG(result ? "Authenticated" : "Not authenticated");
    
    Settings::firewall_map_type::iterator firewall_it = firewall_map->find(user);
    
    bool firewall_result = true;
    
    if(firewall_it != firewall_map->end())
    {
        WOAP2_LOG_TRACE(std::string("Checking firewall"));
        
        firewall_result = utils::is_address_in_range(client_ip, firewall_it->second.first, firewall_it->second.second);
    
        WOAP2_LOG_DEBUG(firewall_result ? "Firewall passwed" : "Firewalled");
    }
    
    bool auth_result = result && firewall_result;
    
    if(!auth_result)
    {
        if(!result)
        {
            WOAP2_LOG_WARN(std::string("Authentication failed for user ") + user + " with password " + password);
        }
        if(!firewall_result)
        {
            WOAP2_LOG_WARN(std::string("User ") + user + " from " + client_ip + " was firewalled");
        }
    }
    
    WOAP2_LOG_TRACE(" < WoapWorker::authenticate");
    
    return auth_result;
}

void WoapWorker::thread_func(const ssh_session session)
{
    WOAP2_LOG_TRACE(" > WoapWorker::thread_func");
    WOAP2_LOG_DEBUG("Data exchange thread started.");
    
    std::string client_ip = utils::get_client_ip(session);
    
    WOAP2_LOG_INFO(std::string("Connection from ") + client_ip);
    
    // Do SSH key exchange
    WOAP2_LOG_TRACE("SSH key exchange")
    if(ssh_handle_key_exchange(session))
    {
        std::string err_str = ssh_get_error(session);
        WOAP2_LOG_ERROR(std::string("ssh_handle_key_exchange error: ") + err_str);
    }
    
    //************ AUTHENTICATE PHASE ************
    WOAP2_LOG_TRACE("SSH auth phase");
    bool auth = false;
    bool auth_failed = false;
    
    ssh_message message;
    
    do
    {
        message = ssh_message_get(session);
        
        if(!message)
        {
            break;
        }
        
        switch(ssh_message_type(message))
        {
            case SSH_REQUEST_AUTH:
            {
                WOAP2_LOG_TRACE("Auth requested");
                switch(ssh_message_subtype(message))
                {
                    case SSH_AUTH_METHOD_PASSWORD:
                    {
                        WOAP2_LOG_TRACE("Password method.");
                        if(authenticate(
                                ssh_message_auth_user(message),
                                ssh_message_auth_password(message),
                                client_ip))
                        {
                            auth = true;
                            ssh_message_auth_reply_success(message, 0);
                        }
                        else
                        {
                            auth_failed = true;
                            ssh_message_reply_default(message);
                        }
                        break;
                    }
                    case SSH_AUTH_METHOD_NONE:
                    default:
                    {
                        WOAP2_LOG_TRACE("Other method. Not supported.");
                        ssh_message_auth_set_methods(message, SSH_AUTH_METHOD_PASSWORD);
                        ssh_message_reply_default(message);
                        break;
                    }
                }
                break;
            }
            default:
            {
                WOAP2_LOG_TRACE("Not SSH_REQUEST_AUTH message type. Unexpected.");
                ssh_message_reply_default(message);
                break;
            }
        }
        
        ssh_message_free(message);
    } while(message && !auth && !auth_failed);
    
    if(!auth)
    {
        std::string err_str = ssh_get_error(session);
    
        ssh_disconnect(session);
    
        if(!auth_failed)
        {
            WOAP2_LOG_ERROR(std::string("Auth phase error: ") + err_str);
        }
    
        WOAP2_LOG_DEBUG("Data exchange thread exit.");
        WOAP2_LOG_TRACE(" < WoapWorker::thread_func");
        return;
    }
    
    //************ SESSION CHANNEL OPEN PHASE ************
    
    WOAP2_LOG_TRACE("SSH session channel open phase");
    ssh_channel chan = 0;
    
    do
    {
        message = ssh_message_get(session);
        if(message)
        {
            switch(ssh_message_type(message))
            {
                case SSH_REQUEST_CHANNEL_OPEN:
                {
                    WOAP2_LOG_TRACE("Channel open requested.");
                    if(ssh_message_subtype(message) == SSH_CHANNEL_SESSION)
                    {
                        WOAP2_LOG_TRACE("Session channel requested.");
                        chan = ssh_message_channel_request_open_reply_accept(message);
                        break;
                    }
                }
                default:
                {
                    WOAP2_LOG_TRACE(
                            "Not SSH_REQUEST_CHANNEL_OPEN message type or not SSH_CHANNEL_SESSION subtype. Unexpected.")
                    ssh_message_reply_default(message);
                    break;
                }
            }
            ssh_message_free(message);
        }
    } while(message && !chan);
    
    if(!chan)
    {
        std::string err_str = ssh_get_error(session);
        ssh_disconnect(session);
    
        WOAP2_LOG_ERROR(std::string("SSH session channel open phase error: ") + err_str);
    
        WOAP2_LOG_DEBUG("Data exchange thread exit.");
        WOAP2_LOG_TRACE(" < WoapWorker::thread_func");
        
        return;
    }
    
    //************ SHELL AND PTY CHANNELS OPEN PHASE ************
    WOAP2_LOG_TRACE("SSH shell and pty phase");
    bool shell = false;
    
    do
    {
        message = ssh_message_get(session);
        if(message)
        {
            switch(ssh_message_type(message))
            {
                case SSH_REQUEST_CHANNEL:
                {
                    WOAP2_LOG_TRACE("SSH channel requested.");
                    int message_subtype = ssh_message_subtype(message);
                    if(message_subtype == SSH_CHANNEL_REQUEST_SHELL)
                    {
                        WOAP2_LOG_TRACE("SSH shell requested");
                        shell = true;
                        ssh_message_channel_request_reply_success(message);
                        break;
                    }
                    else if(message_subtype == SSH_CHANNEL_REQUEST_PTY)
                    {
                        WOAP2_LOG_TRACE("SSH pty requested");
                        ssh_message_channel_request_reply_success(message);
                        break;
                    }
                }
                default:
                {
                    WOAP2_LOG_TRACE(
                            "Not SSH_REQUEST_CHANNEL message type or not SSH_CHANNEL_REQUEST_SHELL or SSH_CHANNEL_REQUEST_PTY subtype. Unexpected.")
                    ssh_message_reply_default(message);
                    break;
                }
            }
            ssh_message_free(message);
        }
    } while(message && !shell);
    
    if(!shell)
    {
        std::string err_str = ssh_get_error(session);
        ssh_disconnect(session);
    
        WOAP2_LOG_ERROR(std::string("SSH shell and pty phase error: ") + err_str);
    
        WOAP2_LOG_DEBUG("Data exchange thread exit.");
        WOAP2_LOG_TRACE(" < WoapWorker::thread_func");
        
        return;
    }
    
    WOAP2_LOG_DEBUG("SSH connection established.");
    
    //************ ACTUAL DATA EXCHANGE ************

    int read = 0;
    char buf[2049];
    
    // Login
    std::unique_ptr<commands::CommandBase> cmd = commands_factory.get_login_command();
    
    std::string response = cmd->execute();
    WOAP2_LOG_TRACE(std::string("Sending ") + response);
    ssh_channel_write(chan, response.c_str(), response.length());
    
    do
    {
        memset(buf, 0, sizeof(buf));
        // read the command
        read = ssh_channel_read(chan, buf, 2048, 0);
    
        if(read > 0)
        {
            WOAP2_LOG_TRACE(std::string("Received: ") + buf);
            
            // get the command object
            // buf+1 is to eliminate leading # character.
            cmd = commands_factory.get_command(buf + 1);
    
            // Execute and send result
            response = cmd->execute();
    
            if(response.length() > 0)
            {
                WOAP2_LOG_TRACE(std::string("Sending: ") + response)
                ssh_channel_write(chan, response.c_str(), response.length());
            }
        }
    } while(read > 0);
    
    WOAP2_LOG_DEBUG("ssh_channel_read cycle ended.");
    
    ssh_channel_free(chan);
    
    ssh_disconnect(session);
    ssh_free(session);
    
    WOAP2_LOG_DEBUG("Data exchange thread exit.");
    WOAP2_LOG_TRACE(" < WoapWorker::thread_func");
}

void WoapWorker::listen_thread_func(const ssh_bind sshbind)
{
    WOAP2_LOG_TRACE(" > WoapWorker::listen_thread_func");
    WOAP2_LOG_DEBUG("Listen thread started.");
    
    int accept_res;
    while(1)
    {
        ssh_session session = ssh_new();
    
        WOAP2_LOG_TRACE("Accepting new connection");
        accept_res = ssh_bind_accept(sshbind, session);
        
        if(accept_res == SSH_ERROR)
        {
//            for(ThreadList::iterator it = threads.begin(); it != threads.end(); ++it)
//            {
//                shutdown(ssh_get_fd(*(it->second)), SHUT_RDWR);
//                //ssh_silent_disconnect(*(it->second));
//                it->first->join();
//            }
    
            if(session)
            {
                ssh_free(session);
            }
    
            if(!cancel)
            {
                const char *err_chars = ssh_get_error(sshbind);
                std::string err_str(err_chars);
    
                WOAP2_LOG_ERROR(std::string("ssh_bind_accept error: ") + err_str);
            }
    
            break;
        }
    
        // Add thread collection
        threads.push_back(ThreadPtr(new boost::thread(boost::bind(&WoapWorker::thread_func, this, session))));
    }
    
    WOAP2_LOG_DEBUG("Listen thread exit.");
    WOAP2_LOG_TRACE(" < WoapWorker::listen_thread_func");
}

void WoapWorker::do_work()
{
    WOAP2_LOG_TRACE(" > WoapWorker::do_work");
    WOAP2_LOG_DEBUG("Main thread started.");
    WOAP2_LOG_INFO("woap2-"
                     Proj_VERSION_STR
                           " started.");
    
    // create SSH bind object and setup it.
    WOAP2_LOG_TRACE("Setting up ssh_bind.")
    ssh_bind sshbind = ssh_bind_new();
    ssh_bind_options_set(sshbind, SSH_BIND_OPTIONS_DSAKEY, dsa_key.c_str());
    ssh_bind_options_set(sshbind, SSH_BIND_OPTIONS_RSAKEY, rsa_key.c_str());
    ssh_bind_options_set(sshbind, SSH_BIND_OPTIONS_BINDPORT, &port);
    
    // start listening
    WOAP2_LOG_TRACE("Start listening.")
    if(ssh_bind_listen(sshbind) < 0)
    {
        //TODO: add logging;
        std::string err_str = ssh_get_error(sshbind);
        WOAP2_LOG_ERROR(std::string("ssh_bind_listen failed: ") + err_str);
        return;
    }
    
    // start listening thread.
    std::unique_ptr<boost::thread> listen_thread = std::unique_ptr<boost::thread>(
            new boost::thread(boost::bind(&WoapWorker::listen_thread_func, this, sshbind)));
    
    // check for cancellation every 200 msec
    while(!cancel)
    {
        usleep(200000);
        for(ThreadList::iterator it = threads.begin(); it != threads.end(); ++it)
        {
            if((*it)->timed_join(boost::posix_time::seconds(0)))
            {
                it = threads.erase(it);
            }
        }
    }
    
    WOAP2_LOG_TRACE("Shutting down listening socket.");
    // shutdown listen socket. this will stop ssh_bind_accept in the listener thread with SSH_ERROR return code
    shutdown(ssh_bind_get_fd(sshbind), SHUT_RD);
    
    // wait for thread
    listen_thread->join();
    
    ssh_bind_free(sshbind);
    
    WOAP2_LOG_DEBUG("Main thread exit.");
    WOAP2_LOG_TRACE(" < WoapWorker::do_work");
}
